// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// ArfMaker.cc
//
// Authors:  Richard Saxton (VEGA/ESA)
//           ESA XMM SOC
//
// Description: Runs arfgen to produce an ARF
//
#include "string.h"
#include "Tracer.h"
#include "metatask.h"
#include "errstr.h"
#include "param.h"
#include "stringutil.h"

#include "arfgen.h"   // tasks

#include "ArfMaker.h"

ArfMaker::ArfMaker(string spectrumset, string arfset,
              bool withbadpixcorr, bool useodfatt,
              string badpixlocation, bool withrmfset,
              string rmfset, bool extendedsource, 
              bool withdetbins, unsigned xbins, unsigned ybins, 
              bool withsourcepos, string sourcecoords, 
              float sourcex, float sourcey,
              bool keeparf)
{
  PROC;

  /* Set up the parameters for calling arfgen */
    openParameters("arfgen");

  /* First the input parameters */
    setParameter("spectrumset",spectrumset);
    setParameter("arfset",arfset);
    setParameter("withbadpixcorr",withbadpixcorr);
    setParameter("useodfatt",useodfatt);

    if (withrmfset) { // To avoid removing arguments
       errstr<<"Not using the RMF "<<rmfset<<" for ARF creation"
             << Sas::message(Msg::App, Msg::Noisy);
    }
    setParameter("withrmfset",withrmfset);
    setParameter("rmfset",rmfset);       
    setParameter("extendedsource",extendedsource); 
    setParameter("withdetbins",withdetbins); 
    setParameter("detxbins",(int)xbins); 
    setParameter("detybins",(int)ybins); 
    setParameter("withsourcepos",withsourcepos);
    setParameter("sourcecoords",sourcecoords);
    setParameter("sourcex",sourcex);
    setParameter("sourcey",sourcey);

   /* Set the badpixlocation filename, removing any table extension */
    std::vector<std::string> v(breakup(badpixlocation, ":"));
    std::string BadpixSet = v[0];

    errstr<<" Badpix "<<BadpixSet
          << Sas::message(Msg::App, Msg::Verbose);
    
    setParameter("badpixlocation",BadpixSet);

   /* Set the detectormap type depending on source type */
    if (extendedsource) 
    {
       setParameter("detmaptype","flat");
    }
    else{
       setParameter("detmaptype","psf");
    }

  /* Set the other parameters */
    setParameter("withdetbounds",false);
    setParameter("modeleffarea",true);
    setParameter("modelquantumeff",true);
    setParameter("modelfiltertrans",true);
    setParameter("modelee",true);
    setParameter("setbackscale",true); // Set the BACKSCAL value
    setParameter("keeparfset",keeparf); 

  /* Now run arfgen using these parameters */ 
    task(arfgen);
   
  /* Shutdown */
    closeParameters();

  ENDPROC;
}

ArfMaker::~ArfMaker()
{
  PROC;
  ENDPROC;
}

