// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// Module: EregionAnalyse
// Author: Richard Saxton 
//
// Description: Main class for calculating optimum image region
//
// Modifications: See ChangeLog
//

#ifndef EregionAnalyse_H
#define EregionAnalyse_H

/* General classes */
#include <iostream>

/* SAS-specific classes */
#include <Tracer.h>
#include <metatask.h>
#include <errstr.h>
#include <param.h>
#include <Dal.h>
#include <Xmm.h>
#include "CalServer.h"
#include "STime.h"

/* Local Classes */
#include "ImageRegion.h"
#include "CalSetter.h"


/****************************************************
 *
 * class EregionAnalyse
 *
 ***************************************************/

class EregionAnalyse
{
public:

  /* Constructor */
  EregionAnalyse();

  /* Destructor */
  ~EregionAnalyse();

  /* Define data structure */
  struct SOURCE {
      double bkgsub_counts,              // Source counts after bkgnd subt
             bkgsub_counts_error,        // Error on subtracted src counts
             bkgsub_count_rate,          // Bckgnd subtracted count rate
             bkgsub_count_rate_error,    // Error on subtracted count rate
             ee,                         // Applied encircled energy correction
             upplim;                     // Upper limit on source counts
  };

  /* Method to calculate the optimum radius */
  double optimumRadius(double src_cnts, double inradius,
                     double bck_flux,
                     double energy, double offax,
                     double phi);

  /* Method to calculate the optimum ellipse */
  FEllipse * optimumEllipse(double optrad, CoordPair centre, double orient);

  /* Calculates the background subtracted source count rate and error */
  void calcSrcCountRate(ImageRegion * srcreg,
                     ImageRegion * bckreg,
                     double bck_flux,
                     bool backgnd_region_set,
                     double energy, double exposure, double cl);

  /* Calculate the upper limit */
  double calcUpplim(double n, double b, double cl);

  /* Make a copy of the exposure map and add in REF keywords if necessary */
  const string expfile_copy(const string & exposurefile,
                                            const string & imagefile);

//  const string EregionAnalyse::getPostFix();

private:


  /* Member fields */
  SOURCE _source;
  ImageRegion * srcreg;
  ImageRegion * bckreg;
  ImageRegion * expreg;
  Psf *  _psf;
  CalSetter * _calset;
  string  _postfix;

};

#endif

