// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// Eupper.cc
//
// Authors:  Richard Saxton (VEGA/ESA)
//           ESA XMM SOC
//
// Description: Task to calculate the count rate upper limit
//              at a given celestial position within an image
//
//              For multiple images the task can either add the exposure times
//              to find a deeper upper limit from a series of observations or
//              average them to extend an energy range, e.g. add XMM pointed
//              images 4000 and 5000 to get the 2-12 keV range
//
//#include <ctime>
#include <iostream>
#include <fstream>
//gcc4.8.2 portability
#include <unistd.h>


#include "Tracer.h"
#include "errstr.h"
#include "param.h"

#include "Dal.h"
#include "Eupper.h"
#include "UpperLimitSolver.h"
#include "ArrayPixelIterator.h"
#include "SscWcsConvertor.h"

/* Method: Eupper() - Constructor */
Eupper::Eupper()
{
  PROC;

    //cout<<"CLOCK: Starting: "<<clock();
  /* Get the parameters from this command line */
    string test = stringParameter("imageset");
    vector<string> imagefiles = splitString(test);
    vector<string> exposurefiles = splitString(stringParameter("exposuremap"));
    vector<string> bckgndfiles = splitString(stringParameter("bkgimageset"));
    double src_ra = realParameter("srcra");
    double src_dec = realParameter("srcdec");
    double src_rad = realParameter("srcradius");

  // Set background region 
    double bck_ra = src_ra, bck_dec = src_dec, bck_rad_in = src_rad;
 
    if (realParameter("bgdra")) 
       bck_ra = realParameter("bgdra");
    if (realParameter("bgddec"))
       bck_dec = realParameter("bgddec");
    if (realParameter("bgdrinner"))
       bck_rad_in = realParameter("bgdrinner");

    double bck_rad_out = realParameter("bgdrouter");

    double ulsig = realParameter("ulsig");

    //cout<<"Bck: "<<bck_ra<<" "<<bck_dec<<" "<<bck_rad_in<<" "<<bck_rad_out<<endl;

  /* Create variables to hold cumulative counts and exposure */
    double src_counts=0.0, bck_counts=0.0, toterr=0.0, exposure=0.0, eef=0.0,
           area_ratio=0.0; 
    //string obsids=" ",date_min="4000-01-01T00:00:00",date_max="0000-01-01T00:00:00";
    string obsids=" ",date_min="Not Set",date_max="----------";

  /* Loop over each image */
    for (unsigned i=0; i<imagefiles.size(); i++)
    {
      /* Find exposure at the source centre from exposure map if
         present. Skip image if exposure=0 */
       double expbit = getExposure(exposurefiles[i], src_ra,
                                    src_dec, 1.0);
       if (expbit>0) {

         /* Get a pixel iterator for the image */
          ImageData srcImage = readImage(imagefiles[i]);
  
         /* Calculate the counts in the source region */
          COUNTS source = counts(srcImage, src_ra, src_dec, 0.0, src_rad);
          src_counts += source.counts;
    
         /* Process the background information: If a background map has been 
         supplied use this in conjunciton with the source spatial expression,
         If a map hasn´t been set use the source image with the bckgnd region */
          COUNTS bck;
          if (bckgndfiles[0] != "NotSet") {
             ImageData bgdImage = readImage(bckgndfiles[i]);
             bck = counts(bgdImage, src_ra, src_dec, 0.0, src_rad);
          } else {
             bck = counts(srcImage, bck_ra, bck_dec, bck_rad_in, bck_rad_out);
          }

         /* Find ratio of source to background region areas and normalise background */
          area_ratio = source.area / bck.area;
          bck_counts += bck.counts * area_ratio;

         /* Find encircled energy factor for source region 
            Implicitly assumes that eef is the same for each image */
          eef = realParameter("eef");  
          if (! booleanParameter("witheef"))   // calculate if not set as paremeter
             eef = getEEF(srcImage, src_rad, source.offaxAngle);

      /* Calculate the error on the bckgnd subtracted count rate */
      // ? Is it correct to just add here - should revisit the non-bayesian calc //
          toterr += sqrt(source.counts + bck.counts
                                       * area_ratio*area_ratio) / eef;

         /* Add on the exposure for this image */
          if (exposurefiles[i] != "NotSet") {
             exposure += expbit;
          } else {
             // Calculate from exposure of image x the vignetting factor
             exposure += srcImage.exposure * 
                                vignetFactor(srcImage, source.offaxAngle);
          }

       //cout<<"Toterr: "<<toterr<<" "<<source.counts<<" "<<bck.counts<<" "<<area_ratio<<" "<<eef<<endl;
      /* Produce strings for obsid(s) and min/max observation dates*/
          obsids = obsids + " " + srcImage.obsid;
          date_min=min(date_min,srcImage.start_date);
          date_max=max(date_max,srcImage.end_date);
       }
    }

  /* Use the average exposure if required. This option would typically be used 
     for joining images from different energy bands */
    if (stringParameter("expmode") == "average")
           exposure = exposure / imagefiles.size();

  /* Calculate upper limit */
    double upplim = calcUpperlimit(exposure, src_counts,
                              bck_counts, toterr, ulsig, eef);

  /* Output useful debug if required */
    //errstr<< "Source/back counts: "<<source.counts<<" "<<bck.counts<<endl
    errstr<< "Source/back counts: "<<src_counts<<" "<<bck_counts<<endl
        << "Area ratio: "<<area_ratio<<endl
        << "Exposure: "<<exposure<<endl
        << "EEF: "<<eef
        << Sas::message(Msg::App, Msg::Verbose);

  /* Calculate the bckgnd subtracted count rate and error */
    double bkg_sub_crate = (src_counts - bck_counts) / eef / exposure;
    double bkg_sub_crate_err = toterr / exposure;

  /* Output the result */
    ostringstream outs;
    outs.precision(8);

    if (booleanParameter("details")) {
       outs<<"Source/back counts: "<<src_counts<<" "<<bck_counts<<endl
           <<"Area ratio: "<<area_ratio<<endl
           << "Exposure: "<<exposure<<endl
           << "EEF: "<<eef<<endl
           << "Obsid: "<<obsids<<endl
           << "Dates: "<<date_min<<" "<<date_max<<endl
           << "Bckgnd subtracted source c/r: "<<bkg_sub_crate
           << " +/- "<<bkg_sub_crate_err<<" c/s" <<endl;
    }
    outs<<" Statistical upper limit c/r: "<<upplim<<" c/s" <<endl;

    if (booleanParameter("withoutputfile")) {

      /* Open a stream to an output text file */
       ofstream outfile;
       string textfile=stringParameter("output");
       outfile.open(textfile.c_str());
       outfile<<outs.str();
       outfile.close();
    
    } else { // Write to standard error

       errstr<<outs.str()
             <<Sas::message(Msg::App, Msg::Always);
    }
  // } REMOVE

  /* Set output parameters */

  ENDPROC;
}

/* Destructor */
Eupper:: ~Eupper()
{
  PROC;

    // CLOSE EVERYTHING DOWN PROPERLY
    //delete srcreg;            srcreg=0;

  ENDPROC;
};

/* Eupper::calcUpplim - calculates the upper limit for total counts
,n, background counts, b and confidence limit, cl */
double Eupper::calcUpplim(double n, double b, double cl)
{
   UpperLimitSolver u = UpperLimitSolver(static_cast<long int>(n),b);
   std::pair<double, double> limits = u.bayesian(cl);
   //double lower = limits.first, upper=limits.second;
   double upper=limits.second;

   //cout<<"LIMITS: "<<lower<<" "<<upper<<endl;

   RETURN upper;
}

Eupper::ImageData Eupper::readImage(const string & imagefile)
{
 /* Open the input image as a dataset */
    ImageData im;

    im.imageset = dataSetServer->open( imagefile, DataSet::Read );
    
   // Set up the WCS matrix - this is used internally within SSCLIB
   // conversion routines
    SscWcsConvertor::setWcs(imagefile);

  /* Check whether image axes are inverted */
 //   bool invert_axes = checkaxes(array);

  /* Make a pixel iterator from the input image */
  /* SHOULD handle inverted axes HERE !! */
    string iname = imagefile;
    if (imagefile.find(":") == string::npos)
          iname = iname+":";

    im.image = new ArrayPixelIterator(iname);

  // Initialise the reading of this array
    im.image->first();

  /* Find the instrument and telescope */
    im.instrument = getAttfromSetorArr(im.imageset, imagefile, "INSTRUME");
    string telescope = getAttfromSetorArr(im.imageset, imagefile, "TELESCOP");

    // Combine instr and telescope to avoid clashes between ROSAT and Einstein
    // if instrument is HRI
    if (im.instrument != "") {
       if (telescope != "") {
          if (telescope.substr(0,5)=="ROSAT" && im.instrument.substr(0,3)=="HRI")
             im.instrument="ROSAT-HRI";
          if (telescope.substr(0,8)=="EINSTEIN" && im.instrument.substr(0,3)=="HRI")
             im.instrument="EINSTEIN-HRI";
       }
    } else {
       errstr<<"Attribute INSTRUME not found in image header"
                   << Sas::warning("InstrumentNotFound");
    }

    if (im.imageset->hasAttribute("OBS_MODE")) 
       im.obsmode = im.imageset->attribute("OBS_MODE")->asString();
    else
       errstr<<"Attribute OBS_MODE not found in image header"
                   << Sas::warning("ObsModeNotFound");

  /* Attempt to get the observation id from the main array */
    if (im.imageset->hasAttribute("OBS_ID"))
       im.obsid = im.imageset->attribute("OBS_ID")->asString();
    else if (im.instrument.substr(0,3) == "CMA") // EXOSAT-LE
       // Get the obsid from the filename !
       im.obsid = imagefile.substr(imagefile.find("CMA")-5,5);
    else if (im.instrument.substr(0,3) == "IPC") // Einstein-IPC
       im.obsid = im.imageset->attribute("OBJECT")->asString();
    else
       im.obsid = "----------";

  /* Attempt to get the observation dates from the main array */
    if (im.imageset->hasAttribute("DATE-OBS")) 
       im.start_date = im.imageset->attribute("DATE-OBS")->asString();
    else if (im.imageset->hasAttribute("DATE_OBS"))
       im.start_date = im.imageset->attribute("DATE_OBS")->asString();
    else
       im.start_date = "----------";

    if (im.imageset->hasAttribute("DATE-END"))
       im.end_date = im.imageset->attribute("DATE-END")->asString();
    else if (im.imageset->hasAttribute("DATE_END"))
       im.end_date = im.imageset->attribute("DATE_END")->asString();
    else if (im.imageset->hasAttribute("DATE_STP"))
       im.end_date = im.imageset->attribute("DATE_STP")->asString();
    else
       im.end_date = "----------";

    // If date has the time but without a 'T' then add it
    // e.g. 1978-12-13 02:07:18.884000 -> 1978-12-13T02:07:18.884000
    if (im.start_date.find("T") == string::npos &&
        im.start_date.find(":") != string::npos)
          std::replace(im.start_date.begin(), im.start_date.end(), ' ', 'T'); 

    if (im.end_date.find("T") == string::npos && 
        im.end_date.find(":") != string::npos)
          std::replace(im.end_date.begin(), im.end_date.end(), ' ', 'T');

    // If date doesnt contain the time and if there is a time keyword
    // then use it
    if (im.start_date.find("T") == string::npos && 
        im.imageset->hasAttribute("TIME-OBS")) 
        im.start_date = im.start_date + "T" + 
               im.imageset->attribute("TIME-OBS")->asString();       

    if (im.end_date.find("T") == string::npos && 
        im.imageset->hasAttribute("TIME-END")) 
        im.end_date = im.end_date + "T" + 
               im.imageset->attribute("TIME-END")->asString();       

  /* Get the exposure from the main array */
    Array * array;
    if (imagefile.find(":") != string::npos) {
       array = im.imageset->array(imagefile);
    } else {
       array = im.imageset->array(imagefile+":");
    }

    // Look for EXPOSURE or ONTIME keyword
    if (array->hasAttribute("EXPOSURE")) {
       im.exposure = array->attribute("EXPOSURE")->asReal();
    } else if (im.imageset->hasAttribute("EXPOSURE")) {
       im.exposure = im.imageset->attribute("EXPOSURE")->asReal();
    } else if (array->hasAttribute("ONTIME")) {
       im.exposure = array->attribute("ONTIME")->asReal();
    } else if (im.imageset->hasAttribute("ONTIME")) {
       im.exposure = im.imageset->attribute("ONTIME")->asReal();
    } else if (im.imageset->hasAttribute("TIME_LIV")) {
       im.exposure = im.imageset->attribute("TIME_LIV")->asReal();
    } else {
       im.exposure = 1.0;
    }

    // Einstein-HRI uses a different keyword but also has EXPOSURE
    if (im.instrument.substr(0,3) == "HRI" && im.imageset->hasAttribute("TIME_NET"))
       im.exposure = im.imageset->attribute("TIME_NET")->asReal();
 
   // If this is a detector image then we also need to set the 
   // spacraft and instrument attitude matrices
    if (im.imageset->hasAttribute("CTYPE1") ) {
       if (im.imageset->attribute("CTYPE1")->asString() == "DETX") {
          SscWcsConvertor::setAtt(imagefile); } }
    else if (array->hasAttribute("CTYPE1") ) {
       if (array->attribute("CTYPE1")->asString() == "DETX" ) {
          SscWcsConvertor::setAtt(imagefile); }}
 
  /* Return */
    return im;
}


/* Method - counts() - returns the total counts in the image region */
Eupper::COUNTS Eupper::counts(ImageData imdat, double ra, double dec,
                      double rad_inner, double rad_outer)
{
  PROC;

     Eupper::COUNTS result;

    // Find image pixel index which contains this RA, DEC
     PixelElement pixel = findPixel(imdat, ra, dec);

     int xcen = pixel.xIndex(), ycen = pixel.yIndex();

    // Calculate width of region in pixels
    // Ensure pixel width is in units of degrees
     double convtodeg=1.0;
     if (imdat.image->wcsconvertor()->xType().find("DET") != string::npos)
        convtodeg=72000.0;
 
     double xhwid = rad_outer / (abs(pixel.xWidth()) / convtodeg);
     double yhwid = rad_outer / (pixel.yWidth() / convtodeg);

    // Find offset from the image centre in arcmins
     double xoff = (imdat.image->xbins()/2.0 - xcen) * 
                       abs(pixel.xWidth()) / convtodeg,
            yoff = (imdat.image->ybins()/2.0 - ycen) *
                       abs(pixel.yWidth()) / convtodeg;
       
     result.offaxAngle = sqrt(xoff*xoff+yoff*yoff) * 60.;

     //cout<<"Pixel index: "<<xcen<<" "<<ycen<<" "<<xhwid
       //  <<" "<<yhwid<<endl;

     double rout_pix = xhwid, 
            rinn_pix = rad_inner / (abs(pixel.xWidth()) / convtodeg);

    // Find X and Y ranges of pixels
     int ixmin=pixel.xIndex() - INT(xhwid) - 1;
     int ixmax=pixel.xIndex() + INT(xhwid) + 1;
     int iymin=pixel.yIndex() - INT(yhwid) - 1;
     int iymax=pixel.yIndex() + INT(yhwid) + 1; 

     int xbins = imdat.image->xbins();
     int ybins = imdat.image->ybins();

     ixmin = min( max(1,ixmin), (xbins));
     iymin = min( max(1,iymin), (ybins));
     ixmax = min( max(1,ixmax), (xbins));
     iymax = min( max(1,iymax), (ybins));

     double total=0.0, offset=0.0;
     int npix_in_area=0;

     //cout<<"CLOCK:Iterating "<<clock()<<endl;
     //cout<<"Iterating over pixels: "<<ixmin<<" "<<ixmax<<" "<<iymin<<" "<<iymax<<endl;

   // Loop over the X,Y range which contains the region
     for (int iy = iymin; iy<=iymax; iy++)
     {
        for (int ix = ixmin; ix<=ixmax; ix++)
        {
        //   imdat.image->setIndex( (iy-1) * xbins + ix -1 );

          // Check if this pixel is within annulus
           offset = sqrt(float( ( xcen - ix )*( xcen - ix )
                   + ( ycen - iy )*( ycen - iy ) ));
           
           if (offset <= rout_pix && offset>= rinn_pix) {
              //total = total+pixel.value();
              unsigned long index = (iy-1) * xbins + ix -1;
              total = total+imdat.image->valueAt(index);
              npix_in_area++;
	      //cout<<"DEBUG: "<<imdat.image->valueAt(index)<<" "<<ix<<" "<<iy<<endl;
           }
           
        }
     }
     result.counts = total;
     result.area = double(npix_in_area);

     //cout<<"CLOCK:Stopped Iterating "<<clock();
     //cout<<"Counts are: "<<result.counts<<" in area "<<result.area
       //  <<" at offax angle "<<result.offaxAngle<<endl;

  RETURN result;
}

double Eupper::getExposure(const string &exposurefile, double src_ra,
                              double src_dec, double defaultExp) 
{   
  PROC;

    double exposure = 0.0;
    if (exposurefile != "NotSet") {

      // Open the exposure file and get the data array
       ImageData expdata = readImage(exposurefile);

      // Get the pixel which contains the source centre
       PixelElement pixel = findPixel(expdata, src_ra, src_dec);

       exposure = pixel.value();

    } else {
       exposure = defaultExp;
    }

    return exposure;
}

double Eupper::calcUpperlimit(double exposure, double src_counts, 
                              double bck_counts, double toterr,
                              double cl, double eef) 
{
  // bck_counts is the background rate normalised to the source region area

   double upplim = 0.0;

   if (exposure > 0.0) {  // Leave upper limit as 0 if exposure <= 0
     if (src_counts < 80.0) { // Use Bayesian statistics
     //if (src_counts < 800.0) { // Use Bayesian statistics
        upplim = calcUpplim(src_counts, bck_counts, cl);
     } else {// Use source counts + 2 x error

    /* Need to convert between confidence interval and sigma.
       Can't find the equation right now - so hardcode for 1,2 and 3-sigma */
       // Formula is Fraction(n) = 0.5*(1+erf(1/sqrt(2)))
       double fac=0.0;
       if ( cl>0.67 && cl<0.69) // 1-sigma has 68.2689492%
          fac=1.0;
       else if ( cl>0.98) // Note: according to Wikipedia 3-sigma has 99.7300204%
          fac=3.0;
       else // 2-sigma has 95.4499%
          fac=2.0;

       //upplim = ((src_counts - bck_counts) + fac * ( toterr * eef ));
       upplim = (max((src_counts - bck_counts),0.0) + fac * ( toterr * eef ));
   //   cout<<"DBG: "<<toterr<<" "<<eef<<" "<<exposure<<" "<<fac<<" "<<upplim<<endl;
     //  cout<<"Comp: "<<upplim<<" "<<calcUpplim(src_counts, bck_counts, cl)<<endl;
     }
   }

 //  cout<<"Test 0 3: "<<calcUpplim(0.0, 3.0, 0.954)<<endl;
//   cout<<"Test 3 0: "<<calcUpplim(3.0, 0.0, 0.954)<<endl;
 //  cout<<"Test 5 5: "<<calcUpplim(5.0, 5.0, 0.954)<<endl;
  // cout<<"Test 10 10: "<<calcUpplim(10.0, 10.0, 0.954)<<endl;
 //  cout<<"Test 50 50: "<<calcUpplim(50.0, 50.0, 0.954)<<endl;
 //  cout<<"Test 100 100: "<<calcUpplim(100.0, 100.0, 0.954)<<endl;
 //  cout<<"Test 1000 1000: "<<calcUpplim(1000.0, 1000.0, 0.954)<<endl;

  // Return the upper limit in c/s corrected for the encircled energy factor
   return upplim / eef / exposure;
}

// Method to calculate the vignetting factor
// in the case where an exposure map is not provided
double Eupper::vignetFactor(ImageData srcImage, double offax) const
{
  // Set default vignetting factor                                                                    
  double vigfac = 1.0;

  // offax - angle in arcminutes                                                                      
  //cout<<"Inst: "<<srcImage.instrument<<endl;

  // XMM EPIC - rough factor for 1 keV                                                                
  if (srcImage.instrument.substr(0,3) == "EPN" ||
      srcImage.instrument.substr(0,4) == "EMOS" ) {
    vigfac = max(0.3,(1.0-0.7*offax/14.));
  }

  // Source: https://heasarc.gsfc.nasa.gov/docs/exosat/exosat_about.html                            
  // "Vignetting in the telescopes reduces the off-axis effective area to 45 %                      
  // of its peak value 1 degree off-axis."                                                          
  else if (srcImage.instrument.substr(0,3) == "CMA") { // EXOSAT-LE                                 
    if (offax <= 60.0) {
      vigfac = 1.0;
    }
    else if (offax > 60.0) {
      vigfac = 0.45;
    }
  }

  // Source: https://heasarc.gsfc.nasa.gov/docs/einstein/RUM_Chapter5.pdf p.18                      
  // Gioia et al. 1990 p.571: Vignetting correction factor = 1 for on-axis and                      
  // increases to 3 for sources off-axis                                                            
  else if (srcImage.instrument == "IPC") { // EINSTEIN-IPC                                          
    if (offax <= 12.0) { // offax [arcmin] smaller or equal then 12 arcmin                          
      vigfac = (-0.0003125 * offax - 0.00825) * offax + 0.997;
    }
    else if (offax > 12.0) { // ANG [arcmin] greater then 12 arcmin                                 
      vigfac = 1.1049 - 0.02136 * offax;
    }
  }

  else if (srcImage.instrument == "PSPC") {
    // from vignet_pspc.fits within EXSAS/MPE, see also Zimmermann et al. (1998) 
    // ftp://ftp.xray.mpe.mpg.de/people/mjf/2RXS/exsas_guide-980708.pdf          
    // see also: https://github.com/olekoenig/upperlimittools/blob/master/code/m\
    // iscellaneous/satellites/rosat/rosat_PSPC_vign.py                                 
    // IMPORTANT NOTE: We assume a constant energy of 1keV. However              
    // the vignetting is energy dependent! This is NOT taken into account!       
    // The values represent channel 63 (~1keV) 
    if ( abs(offax - 2.5) < 2.5 ) { // meaning 0<offax<=5arcmin                  
      vigfac = 1.0;                                                              
    } else if ( abs(offax - 7.5) < 2.5 ) { // 5<offax<=10arcmin                  
      vigfac = 0.987363;                                                         
    } else if ( abs(offax - 12.5) < 2.5 ) { // 10<offax<=15arcmin                
      vigfac = 0.9646626;                                                        
    } else if ( abs(offax - 17.5) < 2.5 ) { // 15<offax<=20arcmin                
      vigfac = 0.9052803;                                                        
    } else if ( abs(offax - 22.5) < 2.5 ) { // 20<offax<=25arcmin                
      vigfac = 0.860131;                                                         
    } else if ( abs(offax - 27.5) < 2.5 ) { // 25<offax<=30arcmin                
      vigfac = 0.8133605;                                                        
    } else if ( abs(offax - 32.5) < 2.5 ) { // 30<offax<=35arcmin                
      vigfac = 0.7549381;                                                        
    } else if ( abs(offax - 37.5) < 2.5 ) { // 35<offax<=40arcmin                
      vigfac = 0.7018803;                                                        
    } else if ( abs(offax - 42.5) < 2.5 ) { // 40<offax<=45arcmin                
      vigfac = 0.6589034;                                                        
    } else if ( abs(offax - 47.5) < 2.5 ) { // 45<offax<=50arcmin                
      vigfac = 0.6036828;                                                        
    } else if ( abs(offax - 52.5) < 2.5 ) { // 50<offax<=55arcmin                
      vigfac = 0.5471003;                                                        
    } else if ( abs(offax - 56.25) < 1.25 ) { // 55.0 < offax <= 57.5            
      vigfac = 0.435174;                                                         
    } else if ( abs(offax - 58.75) < 1.25 ) { // meaning 57.5 < offax <= 60.0    
      vigfac = 0.2079527;                                                        
    } else {                                                                     
      vigfac = 0.03976972;                                                       
    }                                                                            
  }
  else if (srcImage.instrument == "ROSAT-HRI") {
    // Info: run https://github.com/olekoenig/upperlimittools/blob/master/code/m\
    // iscellaneous/satellites/rosat/rosat_HRI_vign.py                                  
    // These values are from the table vignet_hri.fits within EXSAS/MPE          
    // The vignetting is a step function in this case!  
    if (offax == 0.0 ) { // offax [arcmin]                                       
      vigfac = 1.0;                                                              
    } else if ( abs(offax - 2.5) < 2.5 ) { // meaning 0.0<offax<=5.0arcmin       
      vigfac = 0.9855056;                                                        
    } else if ( abs(offax - 7.5) < 2.5 ) { // meaning 5.0<offax<=10.0            
      vigfac = 0.9561262;                                                        
    } else if ( abs(offax - 12.5) < 2.5 ) { // meaning 10.0<offax<=15.0          
      vigfac = 0.9118617;                                                        
    } else if ( abs(offax - 17.5) < 2.5 ) { // meaning 15.0<offax<=20.0          
      vigfac = 0.8527120;                                                        
    } else { // offax > 20.0arcmin                                               
      vigfac = 0.7786772;                                                        
    }   
  }

  //cout<<"Vigfac: "<<vigfac<<endl;
  return vigfac;
}


double Eupper::getEEF(ImageData srcImage, double src_rad_deg, double offax) const
{
  // Hardwire these correction factors to allow the code to run on any image
  // and for any instrument
    double tol=1.0; // arcsecs
    double eef=1.0; // full coverage by default
    double src_rad = src_rad_deg * 3600.0; // In arcseconds
 
    //cout<<"Src rad: "<<src_rad<<endl;

  // EPIC - values are for EPN
    if (srcImage.instrument.substr(0,3) == "EPN" || 
        srcImage.instrument.substr(0,4) == "EMOS" ) {

       if (srcImage.obsmode.substr(0,4) == "SLEW") {
          if ( abs(src_rad - 30.0) < tol ) { 
             eef=0.845;  // This is the slew PSF for 1.5 keV - not very E-dept.
          } else if ( abs(src_rad - 45.0) < tol ) {
             eef=0.907;
          } else if ( abs(src_rad - 60.0) < tol ) {
             eef=0.936;
          }
       } else { // POINTING we assume - set for 6' off-axis, otherwise slow.  
          if ( abs(src_rad - 30.0) < tol ) { 
             eef=0.876;  // This is the pointed PSF for 1.5 keV - not very E-dep
          } else if ( abs(src_rad - 45.0) < tol ) {
             eef=0.927;
          } else if ( abs(src_rad - 60.0) < tol ) {
             eef=0.951;
          }
       }

    // ROSAT: the PSF is energy-independent from 100-200 arcsecs radius (60-90%)
    // in the RASS.
    } else if (srcImage.instrument.substr(0,4) == "PSPC") { // ROSAT
       if (srcImage.obsmode.substr(0,6) == "SURVEY") {
          if ( abs(src_rad - 30.0) < tol ) { 
             eef=0.2;  // This is TBD
          } else if ( abs(src_rad - 45.0) < tol ) {
             eef=0.35;
          } else if ( abs(src_rad - 60.0) < tol ) {
             eef=0.4;
          } else if ( abs(src_rad - 100.0) < tol ) {
             eef=0.58;
          }
       } else { // Pointed data - VERY position-dependent
	        // ROSAT PSPC Pointed EEF from Max-Planck-Institut fuer
	        // extraterrestrische Physik, EXSAS User's Guide, Zimmermann,
	        // U. et al. 1998, p. 228 (fig. 5.16)
	 
	 // NOTE: We compute the ENERGY-INDEPENDENT EEF values and
	 // assume an energy of 1keV
	 
	 double E = 1.0; // Energy [keV]
	 double eps = offax; // Off-axis angle [arcmin]
	 double p3 = 0.075 * pow(E, 1.43);
	 double r1 = 39.95/E;
	 double r2 = 861.9/E;
	 double alpha = 2.119 + 0.212*E;
	 double R = sqrt(50.61* pow(E,-1.472) + 6.8*pow(E,5.62));
	 double sigma = sqrt(108.7* pow(E,-0.888) + 1.121* pow(E,6) + 0.219*pow(eps,2.848));
	 double p2 = min(pow(10,0.639*E+0.052*pow(E,2)-1.635) * exp(-pow(eps/12,2)/2), 1-p3);
	 double p1 = 1-p3-p2;
	 double M = 1/(log(1+pow(r2/r1,2)) + 1/((alpha-2)*(1+pow(r1/r2,2))) );
	 double M_r;
	 if (src_rad <= r2) {
	   M_r = M * 0.5 * log(1+pow(src_rad/r1,2));
	 } else {
	   M_r = M * ( 0.5 * log(1+pow(src_rad/r1,2)) + (1-pow(src_rad/r2,alpha-2)) / ((alpha-2)*(1+pow(r1/r2,2))));
	 }

	 eef = p1 * (1-exp(-0.5*pow(src_rad/sigma,2))) + p2 * (1-exp(-src_rad/R)) + p3 * M_r;
       }

    } else if (srcImage.instrument == "ROSAT-HRI") { // ROSAT HRI
      // Information from EXSAS User's Guide, Zimmermann,                                                                                                                                                                
      // U. et al. 1998, p. 233, formula 5.19, fig. 5.22
      // See also: GitHub of olekoenig: upperlimittools
      // in code/additional_scripts/satellites/rosat/rosat_HRI_eef.py 
      double E = 1.0; // Energy [keV]
      double eps = offax; // Off-axis angle [arcmin]
      double r1 = 39.95/E;
      double r2 = 861.9/E;
      double alpha = 2.119 + 0.212*E;
      double A1 = 0.9638;
      double A2 = 0.1798;
      double A3 = 0.0009;
      double sigma1 = 2.1858;
      double sigma2 = 3.3 + 0.019*eps - 0.016*pow(eps,2) + 0.0044*pow(eps,3);
      double sigma3 = 31.69;
      double M = 0.04*pow(E,1.43); // according to eq.5.16 <-> eq.5.19 gives different result!?
      double D=1-M;
      double P_M;
      if (src_rad <= r2) {
	P_M = M * 0.5 * log(1+pow(src_rad/r1,2));
      } else {
	P_M = M * ( 0.5 * log(1+pow(r2/r1,2)) + (1-pow(src_rad/r2,alpha-2)) / ((alpha-2)*(1+pow(r1/r2,2))));
      }
      double P_D = (A1*pow(sigma1,2) * (1-exp(-0.5*pow(src_rad/sigma1,2))) + 
		    A2*pow(sigma2,2) * (1-exp(-0.5*pow(src_rad/sigma2,2))) +                                                                                                                                       
		    A3*pow(sigma3,2) * (1-exp(-(1+src_rad/sigma3)*src_rad/sigma3)) ) 
	/ ( A1*pow(sigma1,2) + A2*pow(sigma2,2) + A3*pow(sigma3,2));
      eef = D * P_D + M * P_M;
      
    // EXOSAT LE: the PSF HEW is 50% at 24" radius on-axis and 50% in 4' at
    // 1 degree off-axis     
    // NEED TO INTRODUCE OFF_AXIS ANGLE CONCEPT
    } else if (srcImage.instrument.substr(0,3) == "CMA") { // EXOSAT LE
       if (offax <= 6.0) { // Off-axis angle < 0.1 degs
          if ( abs(src_rad - 24.0) < tol ) { 
             eef=0.5;  
          } else if ( abs(src_rad - 60.0) < tol ) {
             eef=0.64; // To Be Confirmed !!!
          }
       } else { //offax angle > 0.1  (more to be added)
          if ( abs(src_rad - 24.0) < tol ) { 
             eef=0.2;  
          } else if ( abs(src_rad - 60.0) < tol ) {
             eef=0.4; // To Be Confirmed !!!
          }
       } // End of offax angle block

    // Einstein HRI: Small PSF (HEW~4"), EEF~1 at ~1arcmin                                            
    // Formula PSF(r)=2.885*10**(-2) * exp(-r/1.96) + 0.01*exp(-r/12.94)                              
    // and integrate EEF=int_0^upperlim / int_0^5arcmin                                               
    } else if (srcImage.instrument == "EINSTEIN-HRI") { // EINSTEIN (HEAO-2)
      //if (srcImage.obsmode.substr(0,6) == "SURVEY") {
        if ( abs(src_rad - 10.0) < tol ) {
          eef=0.67709655;
        } else if ( abs(src_rad - 20.0) < tol ) {
          eef=0.85186258;
        } else if ( abs(src_rad - 30.0) < tol ) {
          eef=0.93139517;
        } else if ( abs(src_rad - 40.0) < tol ) {
          eef=0.96834833;
        } else if ( abs(src_rad - 50.0) < tol ) {
          eef=0.98539711;
        } else if ( abs(src_rad - 60.0) < tol ) {
          eef=0.99324712;
        }
      //} else {
        //errstr<<"No EEF value for EINSTEIN HRI POINTED data"
          //    << Sas::warning("DataNotSupported");
      //}
    
    // Einstein IPC, has very broad PSF, EEF=0.5 at 5.9arcmin, EEF=0.8 at 9arcmin
    } else if (srcImage.instrument.substr(0,3) == "IPC") {
      //if (srcImage.obsmode.substr(0,6) == "SURVEY") {
        if ( abs(src_rad - 20.0) < tol ) {
          eef=0.00220344;
        } else if ( abs(src_rad - 40.0) < tol ) {
          eef=0.00877944;
        } else if ( abs(src_rad - 60.0) < tol ) { // 1 arcmin                                         
          eef=0.01964931;
        } else if ( abs(src_rad - 80.0) < tol ) {
          eef=0.03465797;
        } else if ( abs(src_rad - 100.0) < tol ) {
          eef=0.05362878;
        } else if ( abs(src_rad - 120.0) < tol ) { // 2 arcmin                                        
          eef=0.0762962;
        } else if ( abs(src_rad - 140.0) < tol ) {
          eef=0.10240134;
        } else if ( abs(src_rad - 160.0) < tol ) {
          eef=0.13158992;
        } else if ( abs(src_rad - 180.0) < tol ) { // 3 arcmin                                        
          eef=0.16352336;
        } else if ( abs(src_rad - 4.0*60.0) < tol ) { // 4 arcmin                                     
          eef=0.27292645;
        } else if ( abs(src_rad - 5.0*60.0) < tol ) { // 5 arcmin                                     
          eef=0.39093195;
        } else if ( abs(src_rad - 6.0*60.0) < tol ) { // 6 arcmin                                     
          eef=0.51056946;
        } else if ( abs(src_rad - 7.0*60.0) < tol ) { // 7 arcmin                                     
          eef=0.62206805;
        } else if ( abs(src_rad - 8.0*60.0) < tol ) { // 8 arcmin                                     
          eef=0.71956421;
        } else if ( abs(src_rad - 9.0*60.0) < tol ) { // 9 arcmin                                     
          eef=0.80003657;
        } else if ( abs(src_rad - 10.0*60.0) < tol ) { // 10 arcmin
          eef=0.86298598;
        }
      //} else {
      //  errstr<<"No EEF value for EINSTEIN IPC POINTED data"
        //      << Sas::warning("DataNotSupported");
     // }


    // ASCA SIS instrument: The EEF values are manually read off
    // fig. 6a in Serlemitsos et al. (1995)
    // Note: The EEF there is normalized to 12arcmin. I think this means
    // that above 12arcmin more than 1xPSF is inside the radius (?).
    //            !!! TO BE CONFIRMED !!!
    // The figure also shows that E-dependence can be neglected.
    // Note2: These values include the PSF of the X-ray telescope (XRT)
    //        The ASCA SIS has no intrinsic PSF so this should be correct
    } else if (srcImage.instrument.substr(0,9) == "SIS0,SIS1") { // ASCA SIS instrument
      //if (srcImage.obsmode.substr(0,6) == "SURVEY") {
      if ( abs(src_rad - 30.0) < tol ) { // 30 arcseconds
          eef=0.07;
      } else if ( abs(src_rad - 60.0) < tol ) { // 1 arcmin
          eef=0.15;
      } else if ( abs(src_rad - 90.0) < tol ) { // 1.5 arcmin
          eef=0.23;
      } else if ( abs(src_rad - 120.0) < tol ) { // 2 arcmin
          eef=0.3;
      } else if ( abs(src_rad - 150.0) < tol ) { // 2.5 arcmin
          eef=0.37;
      } else if ( abs(src_rad - 180.0) < tol ) { // 3 arcmin
          eef=0.47;
      } else if ( abs(src_rad - 4.0*60.0) < tol ) { // 4 arcmin
          eef=0.55;
      } else if ( abs(src_rad - 5.0*60.0) < tol ) { // 5 arcmin
          eef=0.7;
      } else if ( abs(src_rad - 6.0*60.0) < tol ) { // 6 arcmin
          eef=0.78;
      } else if ( abs(src_rad - 7.0*60.0) < tol ) { // 7 arcmin
          eef=0.85;
      } else if ( abs(src_rad - 8.0*60.0) < tol ) { // 8 arcmin
          eef=0.89;
      } else if ( abs(src_rad - 9.0*60.0) < tol ) { // 9 arcmin
          eef=0.94;
      } else if ( abs(src_rad - 10.0*60.0) < tol ) { // 10 arcmin
          eef=0.96;
      } else if ( abs(src_rad - 11.0*60.0) < tol ) { // 11 arcmin
          eef=0.98;
      } else if ( abs(src_rad - 12.0*60.0) < tol ) { // 12 arcmin
          eef=1;
      }
    // IMPORTANT NOTE FOR ASCA GIS: The GIS has an intrinsic PSF
    // See https://heasarc.gsfc.nasa.gov/docs/asca/ao7/appendix_e/node6.html fig.5.3c
    
    // NuSTAR images - instrument is FPMA, FPMB or FPCOMB
    } else if (srcImage.instrument.substr(0,2) == "FP") { // NuSTAR
      if ( abs(src_rad - 29) < tol ) { // 30 arcseconds
          eef=0.5; // 50% - the calibrated value
      } else if ( src_rad < 29.0) { // 1 arcmin
          eef=0.5; // No calibration available - use 50% anyway
          errstr<<"NuSTAR image encircled energy uncalibrated for radius="
                <<src_rad<<" arcsecs, using EEF=0.5"
                << Sas::warning("EEFUnknown");
      } else if ( src_rad > 30.0) { // 1 arcmin
          eef=1.0; // No calibration available - use 100%
          errstr<<"NuSTAR image encircled energy uncalibrated for radius="
                <<src_rad<<" arcsecs, using EEF=1.0"
                << Sas::warning("EEFUnknown");
      }      
    } else {
       errstr<<"Instrument / OBS_MODE combination not recognised, using "
             <<"encircled energy fraction = 1.0"
             << Sas::warning("DataNotSupported");
    }
    return eef;
}

bool Eupper::checkaxes(Array *array) 
{
  PROC;

   /* If the X axis name is "Y" or "DETY" or "DEC" then the image
      is inverted */
    if ( array->attribute("CTYPE1")->asString().find("Y") != string::npos ||
         array->attribute("CTYPE1")->asString().find("DEC") != string::npos )
    {
       ALTRETURNV( true );
    }
    RETURN false;
}

// findPixel - a method to find the pixel which contains a given ra, dec
PixelElement Eupper::findPixel(Eupper::ImageData imdat, double ra, 
                                                   double dec) 
{
   // Input ra, dec should be in degrees - here we convert to radians
   double raR=ra*DEG_TO_RAD, decR=dec*DEG_TO_RAD;

   CoordPair p = SscWcsConvertor::raDecToImagePixels(raR, decR);

   int ix = floor(p.x()+0.5), iy=floor(p.y()+0.5);

   imdat.image->setIndex( (iy-1) * imdat.image->xbins() + ix -1 );
   PixelElement pixel = imdat.image->currentItem();

   return pixel;
}

// Get a string keyword from Set or Array (the xmmsep feature)
const string Eupper::getAttfromSetorArr(DataSet *set, 
                         const string & imagefile, const string & key)
{
    string keyout = "";

    string iname = imagefile;
    if (imagefile.find(":") == string::npos)
          iname = iname+":";

    if (set->hasAttribute(key)) {
       keyout = set->attribute(key)->asString();
    } else if (set->block(iname)->type() == Block::ArrayT) {
       Array *delArr = set->array(iname);
       if (delArr->hasAttribute(key)) {
          keyout = delArr->attribute(key)->asString();
       }
    }

    return keyout;
}

// Split string into words
vector<string> Eupper::splitString(const string &text) {
  vector<string> tokens;
  char sep=' ';
  size_t start = 0, end = 0;
  end = text.find(sep, start);
  while (end != std::string::npos) {
    tokens.push_back(text.substr(start, end - start));
    start = end + 1;
    end = text.find(sep, start);
  }
  tokens.push_back(text.substr(start));
  return tokens;
}
