// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// LightCurveMaker.cc
//
// Authors:  Richard Saxton (VEGA/ESA)
//           ESA XMM SOC
//
// Description: Creates a spectrum using evselect
//
#include <Tracer.h>
#include <metatask.h>
#include <errstr.h>
#include <param.h>
#include <Dal.h>
#include <evselect.h>

#include "LightCurveMaker.h"

LightCurveMaker::LightCurveMaker(string lcurvefile, 
                             DatafileAttributes * timeAttributes,
                             string expression, double timebin) : 
_lcurvefile(lcurvefile),
_timebin(timebin)
{
  PROC;

   /* Get time binning parameters from the etimeget parameter block
      before switching to the evselect parameters */
    bool wtr = booleanParameter("withtimeranges");
    float tmin = realParameter("timemin");
    float tmax = realParameter("timemax");

   /* Set up the parameters for calling evselect */
    openParameters("evselect");

   /* Ideally the selection expression should contain just a source
      spatial region. Everything else is defaulted here */

   /* Set the variable parameters */
    setParameter("rateset",_lcurvefile);
    setParameter("table",timeAttributes->datafile());

   /* Set the pattern and flag selection depending on instrument */
    switch (timeAttributes->instrumentID())
    {
    case Xmm::EMOS1 :
    case Xmm::EMOS2 :
      {
        _pattexp="&&(PATTERN<=12)";
        _flagexp="&&(#XMMEA_EM)";
        break;
      }
    case Xmm::EPN :
      {
        _pattexp="&&(PATTERN<=4)";
        _flagexp="&&(FLAG==0)";
        break;
      }
    default:
      errstr << "Dont recognise the instrument "<<_instr
             << Sas::error("invalidInstrument");
    }

    /* Set the required timebin size */
    setParameter("timebinsize",_timebin);

    /* Add the flag and pattern selection into the evselect expression. 
       If a pattern selection already exists then this will not have 
       any effect with current settings */ 

    setParameter("expression",expression+_pattexp+_flagexp);
       
  /* Set the other parameters */
    setParameter("withfilteredset",false);
    setParameter("keepfilteroutput",false);
    setParameter("flagcolumn","EVFLAG");
    setParameter("flagbit",-1);
    setParameter("filtertype","expression");
    setParameter("writedss",true);          
    setParameter("cleandss",false);          
    setParameter("updateexposure",true);          
    setParameter("filterexposure",true);          
    setParameter("timecolumn","TIME");
    setParameter("withzcolumn",false);
    setParameter("withspectrumset",false);
    setParameter("withspecranges",false);
    setParameter("withrateset",true);
    setParameter("maketimecolumn",true);
    setParameter("makeratecolumn",true);
    setParameter("withhistogramset",false);
    setParameter("energycolumn","PI");

  /* Copy optional time range parameters to evselect */

    setParameter("withtimeranges",wtr);
    setParameter("timemin",tmin);
    setParameter("timemax",tmax);
    
  /* Now run evselect using these parameters */ 
    task(evselect);
   
  ENDPROC;
}

/* Destructor */
LightCurveMaker::~LightCurveMaker()
{
  PROC;
  ENDPROC;
}


