// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// UpperLimitSolver.cc
//
// Authors:  Richard Saxton (VEGA/ESA)
//           ESA XMM SOC
//
// Description: Calculates confidence limits
//
#include<iostream>
#include<cmath>
#include<vector>
#include<utility>

#include "UpperLimitSolver.h"

// Constructor
UpperLimitSolver::UpperLimitSolver(long int _n, double _b)
: B(_b), N(_n)
{
  PROC;

        Const_C();

  ENDPROC;
}

// Destructor
UpperLimitSolver::~UpperLimitSolver()
{
  PROC;
  ENDPROC;
}
/*
 * Bayesian calculation method
 * This is a nearly direct copy of code provided by 
 * Andy Beardmore (bayes_CL.cpp)
 */
std::pair<double, double> UpperLimitSolver::bayesian(const double CL) 
{

  std::pair<double, double> S_range = Bayesian_Confidence_Limits(CL); 


  //cout << "S_min " << S_range.first << "   S_max " << S_range.second << endl;

  RETURN S_range;
}

/** 
 * Calculates the factorial of n.
 * Double's are used to avoid overflow when n > 15L.
 */
double UpperLimitSolver::factorial(const double n) {

  if (n <= 1.0) return 1.0;

  return n * factorial(n - 1.0);

  // for large n could use Gosper's approximation
  // n! = sqrt((2 * n + 1/3)*pi) * pow(n, n) * exp(-n)
}

// The const C in the Kraft etal paper
void UpperLimitSolver::Const_C() {

    double sum = 0.0;

    for (long int n = 0; n <= N; n++) {
      sum += exp(-B) * pow(B, static_cast<double>(n)) / 
	factorial(static_cast<double>(n));
    }

    C = (1.0 / sum);

    //    cout << "C " << C << endl;

}


double UpperLimitSolver::ppdf(const double S) {
    return ( C * exp(-(S + B)) * pow((S + B), static_cast<double>(N)) /
	     factorial(static_cast<double>(N)) );
}

std::pair<double, double> UpperLimitSolver::Bayesian_Confidence_Limits(
						     const double CL) {

  size_t nsteps = 100000;

  double S_low = 0.0;

  double S_step = 0.001;

  std::vector<double> prob(nsteps);

  //double total_prob = 0.0;

  double pmax = 0.0;
  double S_pmax = 0.0;
  unsigned int i_pmax = 0;

  for (unsigned int i = 0; i < nsteps; i ++) {

    double S = S_low + i * S_step;

    prob[i] = ppdf(S);

    if (prob[i] > pmax) {
      pmax = prob[i];
      S_pmax = S;
      i_pmax = i;
    }

  }

  //cout << "Pmax " << pmax  
    //     << " S_pmax " << S_pmax << " i_pmax " << i_pmax << endl;

  unsigned int nlower = i_pmax;
  unsigned int nupper = i_pmax;

  //  if (prob[nlower] == 0.0) nupper++;
  if (nlower == 0) nupper++;
  else if (prob[nlower - 1] >= prob[nupper + 1]) nlower--;
  else nupper++;

  double area = (prob[nlower] + 0.5 * (prob[nupper] - prob[nlower])) * S_step;

      //cerr << "nupper " << nupper << " nlower " << nlower << endl;
      //cerr << "prob[nupper] " << prob[nupper] 
  	// << " prob[nlower] " << prob[nlower] << endl;

      //cerr << "area " << area << endl;

  do {

    if (nlower == 0) {
      nupper++;
      area += ((prob[nupper - 1] + 
		0.5 * (prob[nupper] - prob[nupper - 1])) * S_step);
    }
    else if (prob[nlower - 1] >= prob[nupper + 1]) {
      nlower--;
      area += ((prob[nlower] + 
		0.5 * (prob[nlower + 1] - prob[nlower])) * S_step);
    }
    else {
      nupper++;
      area += ((prob[nupper - 1] + 
		0.5 * (prob[nupper] - prob[nupper - 1])) * S_step);
    }
    
        //cerr << "nupper " << nupper << " nlower " << nlower << endl;
        //cerr << "prob[nupper] " << prob[nupper] 
    	 //<< " prob[nlower] " << prob[nlower] << endl;

        //cerr << "area " << area << endl;

  } while (area < CL);



  double S_min = S_low + nlower * S_step;
  double S_max = S_low + nupper * S_step;


  return pair<double, double>(S_min, S_max);

}

