// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// echeckregion.cc
//
// Authors:  Richard Saxton (VEGA/ESA)
//           ESA XMM SOC
//
// Description: Task which checks whether the centre of a region lies
//              within the observing window and field-of-view
//
#include <Tracer.h>
#include <iostream>
#include <fstream>
#include <errstr.h>
#include <param.h>
#include "Evaluator.h"

// SAS packages
#include "GeometryDataServer.h"
#include "arfgen.h"   // tasks
#include "DatafileAttributes.h"   // tasks
#include "CamCoords.h"
#include "TelCoords.h"
#include "PixCoords.h"
#include "PositionUtils.h"
#include "DssBuilder.h"
#include "RegionExtent.h"

#include "echeckregion.h"

void echeckregion()
{
  PROC;

  /* Set the source, backgnd and arf filenames */
    string sourceset, backset, arfset, outrmfset;    

  /* Get the parameters from this command line */
    string srcexp = stringParameter("srcexp");
    string eventfile = stringParameter("table");

  /* Check that the input expressions are valid */
    if (! Evaluator::isValidExpression(srcexp)) {
       errstr<<"Invalid expression: "<<srcexp
             << Sas::error("InvalidExpression");
    }

  /* Read observation attributes from the event file */
    DatafileAttributes * specAttributes = new DatafileAttributes(eventfile);
    DataSet * evset = dataSetServer->open( eventfile, DataSet::Read );

  /* Get the window mode of the central CCD to exclude sources in
     the central chip later on */
    string centCCDmode = " ";
    if (evset->hasAttribute("SUBMODE"))
        centCCDmode = evset->attribute("SUBMODE")->asString();

  /* Initiate some useful geometry stuff */
    double detx=0.0, dety=0.0;
    CamCoords testpixel(0.0,0.0,0.0);
    PixCoords pixcoord(0,0);
                                                                                
    GeometryDataServer * ct = 0; ct = calServer.getAtom(ct);

  /* Get window object for this observing mode */
   ModeParameters *modeParam = 0; modeParam = calServer.getAtom(modeParam);
   CCDwindow *ccdwindow = new CCDwindow(modeParam->windowX0(),
                                         modeParam->windowY0(),
                        (modeParam->windowX0()+modeParam->windowDX()-1),
                        (modeParam->windowY0()+modeParam->windowDY()-1));
                                                                                
   errstr<<"Set up window of size: "<<modeParam->windowX0()<<" "
        << modeParam->windowY0()<<" "
        <<(modeParam->windowX0()+modeParam->windowDX()-1)<<" "
        <<(modeParam->windowY0()+modeParam->windowDY()-1)
        << Sas::message(Msg::App, Msg::Noisy);

  /* Find the centre of the source region */
   errstr << "Computing source position from region extent.."
          << Sas::message(Msg::App, Msg::Verbose);
                                                                                
   PositionUtils *putils = 
           new PositionUtils(specAttributes->instrumentID());
   const PosWCS *poswcs = new PosWCS(evset, specAttributes->rapnt(),
              specAttributes->decpnt(), specAttributes->papnt(), false);

   DataSubspace dssextra = dssBuilder.getDss(srcexp);
   DataSubspace *dss =  new DataSubspace();

   for(unsigned icomp=0; icomp <dssextra.components(); icomp++){
                                                                                
      errstr<<"Adding component "<<icomp<<" to DSS"
              << Sas::message(Msg::App, Msg::Verbose);
                                                                                
      dss->addComponent(*dssextra.component(icomp));
    }

   const RegionExtent *extent = new RegionExtent(dss, putils,
               specAttributes->datamode(),
               specAttributes->observationDate(), poswcs);

   if (extent->isValid())
        {
          detx = (extent->xmin() + extent->xmax()) / 2.0;
          dety = (extent->ymin() + extent->ymax()) / 2.0;
        }
   else
        {
          errstr
             << "Cannot determine source position from source expression"
             << Sas::error("RegionInvalid");
        }
  
   testpixel = putils->detxy005ToCamCoord2(detx, dety);
   pixcoord  = putils->camCoord2ToPixCoord(testpixel);

  /* Find which CCD this is on and set in CAL, then need to reget the
     GeometryDataServer - otherwise out of fOV check doesn't work */
   Ccd ccd = ct->onCcd(testpixel.x(), testpixel.y());
   if (ccd > 0) {
           calServer.state()->ccd()->set(ccd);
   }
                                                                                
   ct = 0; ct = calServer.getAtom(ct);

  /* Get the number of the central CCD */
   unsigned cent_ccd = calServer.instrument()->focusCcd();
                                                                                
  /* Check for outside FOV */
   CalInt16Vector chipX(0, 1), chipY(0, 1);
   chipX[0] = pixcoord.x();
   chipY[0] = pixcoord.y();
   XmmPixelList chipCoord1 = ct->pixCoord1List(chipX, chipY).toChipCoord();
                                                                                
   //bool infov = chipCoord1.insideFOV()[0];
   bool infov = true;  // Turned off FOV checking to avoid chip gap flagging
                                                                                
  /* Check if pixel is outside an observing window. Normally this only
     applies to the central CCD but all chips are affected in PN large
     window mode. */
   bool insideWindow=true;
                                                                                
   if ( (ccd == cent_ccd || ( calServer.instrument()->id() == Xmm::EPN &&
            calServer.state()->mode()->get() == Xmm::PRIME_LARGE_WINDOW ) )
            && ! ( ccdwindow->isPixelInside(
                        PixelLocation(pixcoord.x(),pixcoord.y()))) )
             {insideWindow=false;}
  
  /* In PN small window mode the other CCDs aren't exposed and so inside
     window should also be false */
   if ( calServer.instrument()->id() == Xmm::EPN &&
            calServer.state()->mode()->get() == Xmm::PRIME_SMALL_WINDOW &&
            ccd != cent_ccd )
             {insideWindow=false;}

  /* In MOS Timing, Burst and W6 mode the central CCD is not used for
     Imaging data, so inside window is false */
   if ( (calServer.instrument()->id() == Xmm::EMOS1 || 
         calServer.instrument()->id() == Xmm::EMOS2) &&
        (centCCDmode=="FastUncompressed" || centCCDmode=="FastCompressed" ||
         centCCDmode=="PrimePartialW6") &&
        (ccd == cent_ccd) ) 
             {insideWindow=false;}

   errstr<<"Centre pixel is: "<<detx<<" "<<dety
         <<"Raw: "<<pixcoord.x()<<" "<<pixcoord.y()
         <<" Inside FOV: "<<infov
         <<" ccd = "<<ccd<<" central ccd = "<<cent_ccd
         <<" inside window: "<<insideWindow
         << Sas::message(Msg::App, Msg::Verbose);
                                                                                
  /* Set the output parameter */
   bool status = infov && insideWindow;
   setParameter("status",status);
																				   string message(" Status of the centre of the input region is: ");
   errstr<< message <<status 
         << Sas::message(Msg::App, Msg::Sparse);
                                                                                
  /* Also output status to a file if required */
   string outfile = stringParameter("output");
   if ( outfile.length() )
    {
      std::ofstream out;
      out.open( outfile.c_str() , std::ios::out);
      out << message << status << endl;
      out.close();
    }
  
  ENDPROC;
}

